#version 430 core
#define MAX_LIGHTS 5

#define AMBIENT_LIGHT 0
#define POINT_LIGHT 1
#define DIRECTIONAL_LIGHT 2
#define SPOT_LIGHT 3

in vec3 fragPosition;
in vec2 fragUV;
in vec3 fragNormal;
in vec3 worldPosition;
in vec3 worldNormal;

out vec4 color;

uniform int numLights;
uniform struct Light {
	vec3 position;
	vec3 intensity;
	float attenuation;
	float coneAngle;
	vec3 coneDirection;
	int type;
	bool castShadows;
	vec2 clip;
	mat4 depthViewProjectionMatrix;
	sampler2DShadow shadowMap;
	samplerCube shadowCube;
} lights[MAX_LIGHTS];
uniform bool receiveIllumination;
uniform bool receiveShadows;
uniform bool isSkybox;
uniform samplerCube skybox;

uniform float shininess;
uniform float reflectivity;
uniform sampler2D reflectivityTexture;

uniform mat4 modelMatrix;
uniform mat4 viewMatrix;
uniform mat4 projectionMatrix;
uniform vec3 cameraPosition;
uniform sampler2D textureBitmap;

float bias = 0.0005;
mat4 biasMatrix = mat4(	0.5, 0.0, 0.0, 0.0,
						0.0, 0.5, 0.0, 0.0,
						0.0, 0.0, 0.5, 0.0,
						0.5, 0.5, 0.5, 1.0);

vec3 viewDir, normal, albedo;
vec4 texColor, shadowCoord;

//taken from https://stackoverflow.com/questions/4200224/random-noise-functions-for-glsl
float rand(vec2 seed){
    return fract(sin(dot(seed.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

float getVisibility(int index, vec3 lightDir){
	bias *= tan(acos(max(0.0,dot(normal, normalize(lightDir)))));
	if (lights[index].castShadows) {
		if (receiveShadows){
			if(lights[index].type == POINT_LIGHT){
				float distPositionLightNormalized = (length(lightDir) - lights[index].clip.x) / (lights[index].clip.y - lights[index].clip.x);
				if(texture(lights[index].shadowCube, -lightDir, bias).r < distPositionLightNormalized - bias)
					return 0.0;
			}else if(lights[index].type == DIRECTIONAL_LIGHT){
				shadowCoord = biasMatrix * lights[index].depthViewProjectionMatrix * modelMatrix * vec4(fragPosition, 1);
				if (texture(lights[index].shadowMap, shadowCoord.xyz, bias)  <  shadowCoord.z + bias)
					return 0.0;
			}else if(lights[index].type == SPOT_LIGHT){
				shadowCoord = biasMatrix * lights[index].depthViewProjectionMatrix * modelMatrix * vec4(fragPosition, 1);
				if (texture(lights[index].shadowMap, shadowCoord.xyz / shadowCoord.w, bias) < ((shadowCoord.z - bias) / shadowCoord.w))
					return 0.0;
			}
		}
	}
	return 1.0;
}

vec3 lightContribution(int index){	
	float reflectivityValue = reflectivity == -1.0 ? texture(reflectivityTexture, fragUV).r : reflectivity;
	vec3 reflectiveColor = pow(texture(skybox, reflect(-viewDir, normal)).rgb, vec3(2.2));
	if(lights[index].type == AMBIENT_LIGHT){
		return 	lights[index].intensity * albedo.rgb * (1.0-reflectivityValue) + reflectivityValue * reflectiveColor;
	}

	float diffuse = 0.0, specular = 0.0, attenuation = 0.0;
	vec3 lightDir = lights[index].position - worldPosition;
	vec3 lightDirNormalized = normalize(lights[index].position - worldPosition);
	
	if(lights[index].type == POINT_LIGHT || lights[index].type == SPOT_LIGHT){
		float dist = length(lights[index].position - worldPosition);
		attenuation = 1.0 / (1.0 + lights[index].attenuation * pow(dist, 2));

		if(lights[index].type == SPOT_LIGHT){
			float lightAngle = degrees(acos(dot(-lightDirNormalized, normalize(lights[index].coneDirection))));
			if(lightAngle > lights[index].coneAngle){
				attenuation = 0.0;
			}
		}
	} else if(lights[index].type == DIRECTIONAL_LIGHT){
		attenuation = 1.0;
	}

	diffuse = max(0.0, dot(normal, lightDirNormalized));

	if(diffuse > 0.0) {
		vec3 halfDir = normalize(lightDir + viewDir);
		float specAngle = max(dot(halfDir, normal), 0.0);
		specular = pow(specAngle, shininess);
	}

	vec3 specularColor = specular * lights[index].intensity * albedo.rgb * attenuation;
	vec3 diffuseColor = diffuse * lights[index].intensity * albedo.rgb * attenuation;

	return getVisibility(index, lightDir) * (diffuseColor + specularColor) * (1.0-reflectivityValue) + reflectivityValue * reflectiveColor;
}

void main()
{
	viewDir = normalize(cameraPosition - worldPosition);
	texColor = texture(textureBitmap, fragUV);
	albedo = pow(texColor.rgb, vec3(2.2));
	normal = normalize(worldNormal);

	if(isSkybox){
		color = texture(skybox, worldPosition);
	}else if(receiveIllumination) {
		vec3 colorCombination = vec3(0); 
		for(int i = 0; i < numLights; i++) {
			colorCombination += lightContribution(i);
		}

		colorCombination = colorCombination / (colorCombination + vec3(1.0));
		colorCombination = pow(colorCombination, vec3(1.0/2.2));  
		color = vec4(colorCombination, texColor.w);
	} else {
		color = texColor;
	}

	//taken from https://www.shadertoy.com/view/MlV3R1
	color.rgb += mix(-0.5/255.0, 0.5/255.0, rand(gl_FragCoord.xy));
}