#version 330 core

in vec2 UV;
out vec4 color;

uniform sampler2D renderedTexture;
uniform vec2 window;

const float gamma_in = 2.2;
const float gamma_out = 4.0;

vec3 lanczos_filter(vec2 texture_size, float sharpness)
{
    texture_size.x *= sharpness;

    vec2 dx = vec2(1.0 / texture_size.x, 0.0);
    vec2 coordinate_pixel = UV * texture_size - vec2(0.5, 0.0);

    vec2 texture_coordinate =(floor(coordinate_pixel) + vec2(0.5, 0.0)) / texture_size;
    vec2 d = fract(coordinate_pixel);

    vec4 c = 3.14 * vec4(d.x + 1.0, d.x, d.x - 1.0, d.x - 2.0);

    	c = max(abs(c), 0.00001);
    c = (2.0 * sin(c) * sin(c / 2.0) / (c * c)) / dot(c, vec4(1.0));

    vec4 column1 = vec4(pow(texture(renderedTexture, texture_coordinate).rgb, vec3(gamma_in)),1.0);
    vec4 column2 = vec4(pow(texture(renderedTexture, texture_coordinate + dx).rgb, vec3(gamma_in)), 1.0);

    return (mat4(column1, column1, column2, column2) * c).rgb;
}

vec3 gaussian_filter(vec2 texture_size)
{
    vec2 dx = vec2(1.0 / texture_size.x, 0.0);
    vec2 dy = vec2(0.0,1.0 / texture_size.y);

    vec2 coordinate_pixel = UV * texture_size;
    vec2 texture_coordinate = (floor(coordinate_pixel) + 0.5) / texture_size;

    vec2 d = (fract(coordinate_pixel) - 0.5) * -1.0;


    mat3 row1 = mat3(pow(texture(renderedTexture,(texture_coordinate - dy) - dx).rgb, vec3(gamma_in)), 
    		pow(texture(renderedTexture, (texture_coordinate - dy)).rgb, vec3(gamma_in)),  
    				pow(texture(renderedTexture,(texture_coordinate - dy) + dx).rgb, vec3(gamma_in)));

    mat3 row2 = mat3(pow(texture(renderedTexture,texture_coordinate - dx).rgb, vec3(gamma_in)), 
					pow(texture(renderedTexture, texture_coordinate).rgb, vec3(gamma_in)),  
    				pow(texture(renderedTexture, texture_coordinate + dx).rgb, vec3(gamma_in)));

    mat3 row3 = mat3(pow(texture(renderedTexture,(texture_coordinate + dy) - dx).rgb, vec3(gamma_in)), 
    				pow(texture(renderedTexture, (texture_coordinate + dy)).rgb, vec3(gamma_in)),  
    			pow(texture(renderedTexture, (texture_coordinate + dy) + dx).rgb, vec3(gamma_in)));

    vec3 blur_w_1 =exp2(pow(vec3(d.x - 1.0,d.x, d.x + 1.0) / 0.5 ,vec3(2)) * -1.0);
    vec3 blur_w_2 =exp2(pow(vec3(d.x - 1.0,d.x, d.x + 1.0) / 0.5 ,vec3(2)) * -1.0);
    vec3 blur_w_3 =exp2(pow(vec3(d.x - 1.0,d.x, d.x + 1.0) / 0.5 ,vec3(2)) * -1.0);

    mat3 column = mat3((row1[0] * blur_w_1.x + row1[1] * blur_w_1.y + row1[2] * blur_w_1.z) / (blur_w_1.x + blur_w_1.y + blur_w_1.z),
                      (row2[0] * blur_w_2.x + row2[1] * blur_w_2.y + row2[2] * blur_w_2.z) / (blur_w_2.x + blur_w_2.y + blur_w_2.z),
                       (row3[0] * blur_w_3.x + row3[1] * blur_w_3.y + row3[2] * blur_w_3.z) / (blur_w_3.x + blur_w_3.y + blur_w_3.z));

    vec3 blur_w_res = exp2(pow(vec3(d.y - 1.0,d.y, d.y + 1.0) / 0.5 ,vec3(2)) * -1.0);

    return (column[0] * blur_w_3.x + column[1] * blur_w_3.y + column[2] * blur_w_3.z) / (blur_w_3.x + blur_w_3.y + blur_w_3.z);
}




vec3 weight_mask(float x, vec2 w_Size)
{
    float i = mod(floor(x * w_Size.x * w_Size.x / (w_Size.x * 3.0)), 2); // mask size



    if (i == 0.0) return mix(vec3(1.0, 0.0, 1.0), vec3(1.0, 0.0, 0.0), 2 - 2.0);
    else if (i == 1.0) return vec3(0.0,1.0,0.0);
    else return vec3(0.0, 0.0, 1.0);
}

vec3 weight_scanline(float x, vec3 c)
{


	// scanline min max
    vec3 x_mul = 2.0/ mix(vec3(0.5), vec3(1.5), c);
    vec3 x_offset = x_mul * 0.5;



    return smoothstep(0.0, 1.0, 1.0 - abs(x * x_mul - x_offset)) * x_offset;
}
 
void main(void)
{

	vec2 w_Size = window;

	vec3 color_soft = lanczos_filter(  w_Size, 1);
    vec3 color_sharp = lanczos_filter( w_Size, 3);

	vec3 color_glow =gaussian_filter( w_Size);
    vec3 color_result = sqrt(color_sharp * color_soft) * 
    					weight_scanline(fract(UV.y * w_Size.y), color_soft);
    color_glow =  clamp(color_glow - color_result, 0.0, 1.0);

    color_result +=color_glow * color_glow * 0.1;
    color_result = mix(color_result,color_result * weight_mask(UV.x, w_Size)*2, 1.8) +color_glow * 0.5; // mask strength
    color_result = pow(color_result * 1.5, vec3(1.0 / gamma_out));
    
    color = vec4(color_result,1.0);

    // blur single pass with crt 
    // not working right now

	float radial_blur = 0.1;
	float radial_bright = 1.0;
	vec2 radial_origin = vec2(0.5,0.5);
	vec2 radial_size = 1/w_Size; 
	vec2 TexCoord = UV;


	vec4 SumColor = vec4(0.0, 0.0, 0.0, 0.0);
	TexCoord += radial_size * 0.5 - radial_origin;
 
	for (int i = 0; i < 10; i++) 
	{
		float scale = 1.0 - radial_blur * (float(i) / 9.0);
		SumColor += texture(renderedTexture, TexCoord * scale + radial_origin);
	}
 
	//color = SumColor / 10.0 * radial_bright;
}